/************************************************************************
 *
 * Basic Interval Arithmetic Subroutines Level 0
 * Modified for generalized intervals and Kaucher arithmetic (2005/06 A. Goldsztejn)
 * ---------------------------------------------
 *
 * Copyright (c) 1993, 1997 Olaf Knueppel
 *
 *
************************************************************************/

#include "BiasK0.h"
#include "BiasRnd.h"
#include "BiasKInt.h"
#include <math.h>
#include <stdio.h>

extern const unsigned short BiasCwRoundUp;
extern const unsigned short BiasCwRoundDown;
extern const unsigned short BiasCwRoundNear;


/*
 * The macro SECURE has been inserted in those cases where problems may
 * occur due to too much compiler optimization (using previous calculated
 * results after a rounding mode switch).
 */
#if 1 /* let's be pessimistic */
#define SECURE                  volatile
#define SECUREREAL(x)           (*((volatile REAL * const) (x)))
#define SECUREBIASKINTERVAL(x)   ((volatile BIASKINTERVAL * const) (x))
#else
#define SECURE
#define SECUREREAL(x)           (*(x))
#define SECUREBIASKINTERVAL(x)   (x)
#endif

#if defined (__BIASINLINEROUNDING__)
VOID BiasKRoundUp (VOID)   { _BiasRoundUp (); }
VOID BiasKRoundDown (VOID) { _BiasRoundDown (); }
VOID BiasKRoundNear (VOID) { _BiasRoundNear (); }
#endif

REAL BiasKEpsilon, BiasKEta, BiasKNaN, BiasKPosInf, BiasKNegInf;

static const char ErrorDivideByZero[] = "Division by Zero";

// BiasK (added for multiplication)
#define IP(px) (BiasKInf(px)>=0)&&(BiasKSup(px)>=0)
#define IZ(px) (BiasKInf(px)<=0)&&(0<=BiasKSup(px))
#define IMP(px) (BiasKInf(px)<=0)&&(BiasKSup(px)<=0)
#define IDZ(px) (BiasKSup(px)<=0)&&(0<=BiasKInf(px))
//

/************************************************************************
*  Implementation of local functions
*************************************************************************/

VOID _BiasKError(const char * const s)
// BiasK
  /**********************************************************************
   *  Report error message and abort
   */
{
  fprintf (stderr, "BIASK: %s\n", (char *) s);
  abort ();
}

/************************************************************************
*  Implementation of global functions
*************************************************************************/

VOID BiasKInit (VOID)
// BiasK
  /**********************************************************************
   *  Initialization
   */
{
  volatile REAL Last, Current, OnePlusCurrent;
  union UREAL { unsigned short ushort[4]; double real; };
#if defined (__BIGENDIAN__)
  static const union UREAL PosInfinity = {{ 0x7FF0, 0x0000, 0x0000, 0x0000 }};
  static const union UREAL NegInfinity = {{ 0xFFF0, 0x0000, 0x0000, 0x0000 }};
  static const union UREAL NaN         = {{ 0x7FFF, 0xFFFF, 0xFFFF, 0xFFFF }};
#else
  static const union UREAL PosInfinity = {{ 0x0000, 0x0000, 0x0000, 0x7FF0 }};
  static const union UREAL NegInfinity = {{ 0x0000, 0x0000, 0x0000, 0xFFF0 }};
  static const union UREAL NaN         = {{ 0xFFFF, 0xFFFF, 0xFFFF, 0x7FFF }};
#endif


  /* Initialize rounding mode switches and set default rounding mode */
  _BiasRoundInit ();
  _BiasRoundNear ();

  /* Calculation of BiasKEpsilon (min { x >= 0 : 1 + x > 1 }) */
  Current = 1.0;
  do {
    Last = Current;
    Current /= 2.0;
    OnePlusCurrent = 1.0 + Current;
  } while (OnePlusCurrent > 1.0);
  BiasKEpsilon = Last;

  /* Calculation of BiasKEta     (min { x > 0 }) */
  Current = Last;
  do {
    Last = Current;
    Current /= 2.0;
  } while (Current > 0.0);
  BiasKEta = Last;

  /* Definition of special floating point numbers (+/-oo, NaN) */
  BiasKPosInf = PosInfinity.real;
  BiasKNegInf = NegInfinity.real;
  BiasKNaN    = NaN.real;
  
#undef EXPIDX
} /* BiasKInit */

REAL BiasKPredR (const REAL * const pa)
// BiasK
  /**********************************************************************
  *  Result: largest possible floating point number with r < a
  */
{
  volatile REAL pr;

  _BiasKRoundDown ();
  pr = *pa - BiasKEta;
  _SetRoundToNearest();
  return pr;
} /* BiasKPredR */

INT BiasKPredI (BIASKINTERVAL * const pR, const BIASKINTERVAL * const pA)
// BiasK
  /**********************************************************************
  *  R = largest possible interval with R.inf > A.inf and R.sup < A.sup
  *  Result: 1, if R is a true interval, 0 otherwise
  */
{
  _BiasKRoundDown ();
  pR->sup = pA->sup - BiasKEta;
  _BiasKRoundUp ();
  pR->inf = pA->inf + BiasKEta;
  _SetRoundToNearest();
  return (pR->inf <= pR->sup);
} /* BiasKPredI */

REAL BiasKSuccR (const REAL * const pa)
// BiasK
  /**********************************************************************
  *  Result: smallest possible floating point number with r > a
  */
{
  volatile REAL pr;

  _BiasKRoundUp ();
  pr = *pa + BiasKEta;
  _SetRoundToNearest();
  return pr;
} /* BiasKSuccR */

VOID BiasKSuccR2I (BIASKINTERVAL * const pR, const REAL * const pa)
// BiasK
  /**********************************************************************
  *  R = smallest interval that contains *pa
  */
{
  _BiasKRoundDown ();
  pR->inf = *pa - BiasKEta;
  _BiasKRoundUp ();
  pR->sup = *pa + BiasKEta;
  _SetRoundToNearest();
} /* BiasKSuccI */

VOID BiasKSuccI (BIASKINTERVAL * const pR, const BIASKINTERVAL * const pA)
// BiasK
  /**********************************************************************
  *  R = smallest possible interval with R.inf < A.inf and R.sup > A.sup
  */
{
  _BiasKRoundDown ();
  pR->inf = pA->inf - BiasKEta;
  _BiasKRoundUp ();
  pR->sup = pA->sup + BiasKEta;
  _SetRoundToNearest();
} /* BiasKSuccI */

VOID BiasKAddRR (BIASKINTERVAL * const pR,
		const REAL * const pa,
                const REAL * const pb)
// BiasK
/**********************************************************************
  *  R = a + b
  */
{
  _BiasKRoundDown ();
  pR->inf = SECUREREAL (pa) + SECUREREAL (pb);
  _BiasKRoundUp ();
  pR->sup = SECUREREAL (pa) + SECUREREAL (pb);
  _SetRoundToNearest();
} /* BiasKAddRR */

VOID BiasKAddRI (BIASKINTERVAL * const pR,
		const REAL * const pa,
                const BIASKINTERVAL * const pB)
// BiasK
  /**********************************************************************
  *  R = a + B
  */
{
  REAL a = *pa;

  _BiasKRoundDown ();
  pR->inf = a + pB->inf;
  _BiasKRoundUp ();
  pR->sup = a + pB->sup;
  _SetRoundToNearest();
} /* BiasKAddRI */

VOID BiasKAddII (BIASKINTERVAL * const pR,
		const BIASKINTERVAL * const pA, const BIASKINTERVAL * const pB)
// BiasK
  /**********************************************************************
  *  R = A + B
  */
{
  _BiasKRoundDown ();
  pR->inf = pA->inf + pB->inf;
  _BiasKRoundUp ();
  pR->sup = pA->sup + pB->sup;
  _SetRoundToNearest();
} /* BiasKAddII */

VOID BiasKSubRR (BIASKINTERVAL * const pR,
		const REAL * const pa,
                const REAL * const pb)
// BiasK
  /**********************************************************************
  *  R = a - b
  */
{
  _BiasKRoundDown ();
  pR->inf = SECUREREAL (pa) - SECUREREAL (pb);
  _BiasKRoundUp ();
  pR->sup = SECUREREAL (pa) - SECUREREAL (pb);
  _SetRoundToNearest();
} /* BiasKSubRR */

VOID BiasKSubRI (BIASKINTERVAL * const pR,
		const REAL * const pa,
                const BIASKINTERVAL * const pB)
// BiasK
  /**********************************************************************
  *  R = a - B
  *  Assertion: R != B
  */
{
  REAL a = *pa;

  _BiasKRoundDown ();
  pR->inf = a - pB->sup;
  _BiasKRoundUp ();
  pR->sup = a - pB->inf;
  _SetRoundToNearest();
} /* BiasKSubRI */

VOID BiasKSubIR (BIASKINTERVAL * const pR,
		const BIASKINTERVAL * const pA,
                const REAL * const pb)
// BiasK
  /**********************************************************************
  *  R = A - b
  */
{
  REAL b = *pb;

  _BiasKRoundDown ();
  pR->inf = pA->inf - b;
  _BiasKRoundUp ();
  pR->sup = pA->sup - b;
  _SetRoundToNearest();
} /* BiasKSubRI */

VOID BiasKSubII (BIASKINTERVAL * const pR,
		const BIASKINTERVAL * const pA, const BIASKINTERVAL * const pB)
// BiasK
  /**********************************************************************
  *  R = A - B
  *  Assertion: R != B
  */
{
  _BiasKRoundDown ();
  pR->inf = pA->inf - pB->sup;
  _BiasKRoundUp ();
  pR->sup = pA->sup - pB->inf;
  _SetRoundToNearest();
} /* BiasKSubII */

VOID BiasKMulRR (BIASKINTERVAL * const pR,
		const REAL * const pa, const REAL * const pb)
// BiasK
  /**********************************************************************
  *  R = a * b
  */
{
  _BiasKRoundDown ();
  pR->inf = SECUREREAL (pa) * SECUREREAL (pb);
  _BiasKRoundUp ();
  pR->sup = SECUREREAL (pa) * SECUREREAL (pb);
  _SetRoundToNearest();
} /* BiasKMulRR */

VOID BiasKMulRI (BIASKINTERVAL * const pR,
		const REAL * const pa, const BIASKINTERVAL * const pB)
// BiasK
  /**********************************************************************
  *  R = a * B
  *  Assertion: R != B
  */
{
  REAL a = SECUREREAL (pa);

  _BiasKRoundDown ();
  if (Sign (SECUREREAL (pa)) == 0) { /* a >= 0 */
    pR->inf = a * pB->inf;
    _BiasKRoundUp();
    pR->sup = a * pB->sup;
  }
  else {                             /* a < 0  */
    pR->inf = a * pB->sup;
    _BiasKRoundUp ();
    pR->sup = a * pB->inf;
  }
  _SetRoundToNearest();
} /* BiasKMulRI */

VOID BiasKMulII (BIASKINTERVAL * const pR,
		const BIASKINTERVAL * const pA,
		const BIASKINTERVAL * const pB)
// BiasK
  /**********************************************************************
  *  R = A * B
  *  Assertion: R != A and R != B
  */
{
  _BiasKRoundDown ();
  if(IP(SECUREBIASKINTERVAL(pA)))
      if(IP(SECUREBIASKINTERVAL(pB)))
            {pR->inf=SECUREBIASKINTERVAL(pA)->inf*SECUREBIASKINTERVAL(pB)->inf ; _BiasKRoundUp(); pR->sup=SECUREBIASKINTERVAL(pA)->sup*SECUREBIASKINTERVAL(pB)->sup;}
      else if (IZ(SECUREBIASKINTERVAL(pB)))
            {pR->inf=SECUREBIASKINTERVAL(pA)->sup*SECUREBIASKINTERVAL(pB)->inf ; _BiasKRoundUp(); pR->sup=SECUREBIASKINTERVAL(pA)->sup*SECUREBIASKINTERVAL(pB)->sup;}
      else if (IMP(SECUREBIASKINTERVAL(pB)))
            {pR->inf=SECUREBIASKINTERVAL(pA)->sup*SECUREBIASKINTERVAL(pB)->inf ; _BiasKRoundUp(); pR->sup=SECUREBIASKINTERVAL(pA)->inf*SECUREBIASKINTERVAL(pB)->sup;}
      else
            {pR->inf=SECUREBIASKINTERVAL(pA)->inf*SECUREBIASKINTERVAL(pB)->inf ; _BiasKRoundUp(); pR->sup=SECUREBIASKINTERVAL(pA)->inf*SECUREBIASKINTERVAL(pB)->sup;}
  else if (IZ(SECUREBIASKINTERVAL(pA)))
      if(IP(SECUREBIASKINTERVAL(pB)))
            {pR->inf=SECUREBIASKINTERVAL(pA)->inf*SECUREBIASKINTERVAL(pB)->sup ; _BiasKRoundUp(); pR->sup=SECUREBIASKINTERVAL(pA)->sup*SECUREBIASKINTERVAL(pB)->sup;}
      else if (IZ(SECUREBIASKINTERVAL(pB)))
            {REAL t1,t2,t3,t4;
            t1=SECUREBIASKINTERVAL(pA)->inf*SECUREBIASKINTERVAL(pB)->sup; t2=SECUREBIASKINTERVAL(pA)->sup*SECUREBIASKINTERVAL(pB)->inf;
            _BiasKRoundUp();
            t3=SECUREBIASKINTERVAL(pA)->inf*SECUREBIASKINTERVAL(pB)->inf; t4=SECUREBIASKINTERVAL(pA)->sup*SECUREBIASKINTERVAL(pB)->sup;
            pR->inf = (t1<t2) ? t1 : t2;
            pR->sup = (t3<t4) ? t4 : t3; }
      else if (IMP(SECUREBIASKINTERVAL(pB)))
            {pR->inf=SECUREBIASKINTERVAL(pA)->sup*SECUREBIASKINTERVAL(pB)->inf ; _BiasKRoundUp(); pR->sup=SECUREBIASKINTERVAL(pA)->inf*SECUREBIASKINTERVAL(pB)->inf;}
      else
            {_BiasKRoundNear() ; pR->inf=0.0 ; pR->sup=0.0;}
  else if (IMP(SECUREBIASKINTERVAL(pA)))
      if(IP(SECUREBIASKINTERVAL(pB)))
            {pR->inf=SECUREBIASKINTERVAL(pA)->inf*SECUREBIASKINTERVAL(pB)->sup ; _BiasKRoundUp(); pR->sup=SECUREBIASKINTERVAL(pA)->sup*SECUREBIASKINTERVAL(pB)->inf;}
      else if (IZ(SECUREBIASKINTERVAL(pB)))
            {pR->inf=SECUREBIASKINTERVAL(pA)->inf*SECUREBIASKINTERVAL(pB)->sup ; _BiasKRoundUp(); pR->sup=SECUREBIASKINTERVAL(pA)->inf*SECUREBIASKINTERVAL(pB)->inf;}
      else if (IMP(SECUREBIASKINTERVAL(pB)))
            {pR->inf=SECUREBIASKINTERVAL(pA)->sup*SECUREBIASKINTERVAL(pB)->sup ; _BiasKRoundUp(); pR->sup=SECUREBIASKINTERVAL(pA)->inf*SECUREBIASKINTERVAL(pB)->inf;}
      else
            {pR->inf=SECUREBIASKINTERVAL(pA)->sup*SECUREBIASKINTERVAL(pB)->sup ; _BiasKRoundUp(); pR->sup=SECUREBIASKINTERVAL(pA)->sup*SECUREBIASKINTERVAL(pB)->inf;}
  else
      if(IP(SECUREBIASKINTERVAL(pB)))
            {pR->inf=SECUREBIASKINTERVAL(pA)->inf*SECUREBIASKINTERVAL(pB)->inf ; _BiasKRoundUp(); pR->sup=SECUREBIASKINTERVAL(pA)->sup*SECUREBIASKINTERVAL(pB)->inf;}
      else if (IZ(SECUREBIASKINTERVAL(pB)))
            {_BiasKRoundNear() ; pR->inf=0.0 ; pR->sup=0.0;}
      else if (IMP(SECUREBIASKINTERVAL(pB)))
            {pR->inf=SECUREBIASKINTERVAL(pA)->sup*SECUREBIASKINTERVAL(pB)->sup ; _BiasKRoundUp(); pR->sup=SECUREBIASKINTERVAL(pA)->inf*SECUREBIASKINTERVAL(pB)->sup;}
      else
            {REAL t1,t2,t3,t4;
            t1=SECUREBIASKINTERVAL(pA)->inf*SECUREBIASKINTERVAL(pB)->sup; t2=SECUREBIASKINTERVAL(pA)->sup*SECUREBIASKINTERVAL(pB)->inf;
            _BiasKRoundUp();
            t3=SECUREBIASKINTERVAL(pA)->inf*SECUREBIASKINTERVAL(pB)->inf; t4=SECUREBIASKINTERVAL(pA)->sup*SECUREBIASKINTERVAL(pB)->sup;
            pR->sup = (t1<t2) ? t1 : t2;
            pR->inf = (t3<t4) ? t4 : t3; }

  _SetRoundToNearest();
} /* BiasKMulII */

VOID BiasKDivRR (BIASKINTERVAL * const pR,
		const REAL * const pa, const REAL * const pb)
// BiasK
  /**********************************************************************
  *  R = a / b
  */
{
#if defined (__BIASKRAISEDIVIDEBYZERO__)
  if (Zero (SECUREREAL (pb))) _BiasKError (ErrorDivideByZero);
#endif
  _BiasKRoundDown ();
  pR->inf = SECUREREAL (pa) / SECUREREAL (pb);
  _BiasKRoundUp ();
  pR->sup = SECUREREAL (pa) / SECUREREAL (pb);
  _SetRoundToNearest();
} /* BiasKDivRR */

VOID BiasKDivRI (BIASKINTERVAL * const pR,
		const REAL * const pa, const BIASKINTERVAL * const pB)
// BiasK
  /**********************************************************************
  *  R = a / B
  */
{
    BIASKINTERVAL T;
    
    BiasKSuccR2I(&T,pa);
    BiasKDivII(pR,&T,pB);
} /* BiasKDivRI */

VOID BiasKDivIR (BIASKINTERVAL * const pR,
		const BIASKINTERVAL * const pA, const REAL * const pb)
// BiasK
  /**********************************************************************
  *  R = A / b
  */
{
    BIASKINTERVAL T;
    
    BiasKSuccR2I(&T,pb);
    BiasKDivII(pR,pA,&T);
} /* BiasKDivIR */

VOID BiasKDivII (BIASKINTERVAL * const pR,
		const BIASKINTERVAL * const pA,
		const BIASKINTERVAL * const pB)
// BiasK
// could be optimized: two roundinds are performed in this version
  /**********************************************************************
  *  R = A / B
  */
{
      if (IZ(pB) || IDZ(pB))
#if defined (__BIASKRAISEDIVIDEBYZERO__)
        _BiasKError (ErrorDivideByZero);
#else
      {
	      pR->sup = pR->inf = BiasKNaN;
      }
#endif
      else
      {
          BIASKINTERVAL T;
          
          T.inf = 1/ ( SECUREBIASKINTERVAL(pB)->sup );
          T.sup = 1/ ( SECUREBIASKINTERVAL(pB)->inf );
	  BiasKSuccI (&T, &T);
          BiasKMulII(pR,pA,&T);
      }        
}

VOID BiasKMacRR (BIASKINTERVAL * const pR,
		const REAL * const pa, const REAL * const pb)
// BiasK
  /**********************************************************************
  *  R += a * b
  */
{
  _BiasKRoundDown ();
  pR->inf += SECUREREAL (pa) * SECUREREAL (pb);
  _BiasKRoundUp ();
  pR->sup += SECUREREAL (pa) * SECUREREAL (pb);
  _SetRoundToNearest();
} /* BiasKMacRR */

VOID BiasKMacRI (BIASKINTERVAL * const pR,
		const REAL * const pa, const BIASKINTERVAL * const pB)
// BiasK
  /**********************************************************************
  *  R += a * B
  *  Assertion: R != B
  */
{
  REAL a = SECUREREAL (pa);

  _BiasKRoundDown ();
  if (Sign (SECUREREAL (pa)) == 0) { /* a >= 0 */
    pR->inf += a * pB->inf;
    _BiasKRoundUp();
    pR->sup += a * pB->sup;
  }
  else {                             /* a < 0  */
    pR->inf += a * pB->sup;
    _BiasKRoundUp ();
    pR->sup += a * pB->inf;
  }
  _SetRoundToNearest();
} /* BiasKMacRI */

VOID BiasKMacII (BIASKINTERVAL * const pR,
		const BIASKINTERVAL * const pA,
		const BIASKINTERVAL * const pB)
// BiasK
  /**********************************************************************
  *  R += A * B
  */
{
  BIASKINTERVAL T;

  BiasKMulII(&T,pA,pB);
  BiasKAddII(pR,pR,&T);
} /* BiasKMacII */

VOID BiasKNeg (BIASKINTERVAL * const pR, const BIASKINTERVAL * const pA)
// BiasK
  /**********************************************************************
  *  R = - A
  *  Assertion: R != A
  */
{
  pR->inf = -(pA->sup);
  pR->sup = -(pA->inf);
} /* BiasKNeg */

VOID BiasKDual (BIASKINTERVAL * const pR, const BIASKINTERVAL * const pA)
// BiasK
  /**********************************************************************
  *  R = dual A
  *  Assertion: R != A
  */
{
  double temp;

  temp = (pA->sup);
  pR->sup = (pA->inf);
  pR->inf = temp;
} /* BiasKDual */
VOID BiasKPro (BIASKINTERVAL * const pR, const BIASKINTERVAL * const pA)
// BiasK
  /**********************************************************************
  *  R = pro A
  *  Assertion: R != A
  */
{
    if (pA->inf<=pA->sup)
    {
        pR->inf = pA->inf;
        pR->sup = pA->sup;
    }
    else
    {
        pR->inf = pA->sup;
        pR->sup = pA->inf;
    }
} /* BiasKPro */
VOID BiasKImp (BIASKINTERVAL * const pR, const BIASKINTERVAL * const pA)
// BiasK OK ( NEW )
  /**********************************************************************
  *  R = imp A
  *  Assertion: R != A
  */
{
    if (pA->inf>=pA->sup)
    {
        pR->inf = pA->inf;
        pR->sup = pA->sup;
    }
    else
    {
        pR->inf = pA->sup;
        pR->sup = pA->inf;
    }
} /* BiasKImp */

REAL BiasKMid (const BIASKINTERVAL * const pA)
// BiasK
  /**********************************************************************
  *  Result: Midpoint(A)
  */
{
  REAL pr;
  _SetRoundToNearest();
  pr = pA->inf + (pA->sup - pA->inf) / 2.0;
  return pr;
} /* BiasKMid */

REAL BiasKDiam (const BIASKINTERVAL * const pA)
// BiasK
  /**********************************************************************
  *  Result: Diameter(A)
  */
{
  REAL pr;

  _SetRoundToNearest();
  pr = pA->sup - pA->inf;
  return pr;
} /* BiasKDiam */

VOID BiasKMidRad (REAL * const pm, REAL * const pr,
		 const BIASKINTERVAL * const pA)
// BiasK
  /**********************************************************************
  *  pm = Midpoint(A), pr = Radius(A)
  */
{
  REAL r;

  _SetRoundToNearest();
  r = (pA->sup - pA->inf) / 2.0;
  *pm = pA->inf + r;
  *pr = r;
} /* BiasKMidRad */

REAL BiasKMig (const BIASKINTERVAL * const pA)
// BiasK (new function)
  /**********************************************************************
  *  Result: Mignitude(A)=Mignitude(proA)
  */
{
  REAL infabs = fabs(pA->inf);
  REAL supabs = fabs(pA->sup);

  if ((pA->inf*pA->sup<=0))
    return 0.0;
  else
    return (infabs < supabs) ? supabs : infabs;
} /* BiasKMignitude */

REAL BiasKAbs (const BIASKINTERVAL * const pA)
// BiasK
  /**********************************************************************
  *  Result: Magnitude(A)=Magnitude(proA)
  */
{
  REAL infabs = fabs(pA->inf);
  REAL supabs = fabs(pA->sup);

  return (infabs < supabs) ? supabs : infabs;
} /* BiasKAbs */

REAL BiasKDistRI (const REAL * const pa, const BIASKINTERVAL * const pB)
// BiasK
  /**********************************************************************
  *  Result: Distance between [a,a] and B
  */
{
  REAL a = *pa;
  REAL infabs = fabs(a - pB->inf);
  REAL supabs = fabs(a - pB->sup);

  return ((infabs < supabs) ? supabs : infabs);
} /* BiasKDistRI */

REAL BiasKDistII (const BIASKINTERVAL * const pA, const BIASKINTERVAL * const pB)
// BiasK
  /**********************************************************************
  *  Result: Distance between A and B
  */
{
  REAL infabs = fabs(pA->inf - pB->inf);
  REAL supabs = fabs(pA->sup - pB->sup);

  return ((infabs < supabs) ? supabs : infabs);
} /* BiasKDistII */

VOID BiasKRR2I (BIASKINTERVAL * const pR,
		 const REAL * const pa, const REAL * const pb)
{
  pR->inf= *pa;
  pR->sup= *pb;
}

VOID BiasKMeetRR (BIASKINTERVAL * const pR,
		 const REAL * const pa, const REAL * const pb)
// BiasK
{
  REAL a = *pa;
  REAL b = *pb;

  if (a > b) { pR->inf = a; pR->sup = b; }
  else       { pR->inf = b; pR->sup = a; }
} /* BiasKMeetRR */

VOID BiasKMeetRI (BIASKINTERVAL * const pR,
		 const REAL * const pa, const BIASKINTERVAL * const pB)
// BiasK
{
  REAL a = *pa;

  pR->inf = (a > pB->inf) ? a : pB->inf;
  pR->sup = (a < pB->sup) ? a : pB->sup;
} /* BiasKMeetRI */

VOID BiasKMeetII (BIASKINTERVAL * const pR,
		 const BIASKINTERVAL * const pA, const BIASKINTERVAL * const pB)
// BiasK
{
  pR->inf = (pA->inf > pB->inf) ? pA->inf : pB->inf;
  pR->sup = (pA->sup < pB->sup) ? pA->sup : pB->sup;
} /* BiasKMeetII */

VOID BiasKJoinRR (BIASKINTERVAL * const pR,
		 const REAL * const pa, const REAL * const pb)
// BiasK
{
  REAL a = *pa;
  REAL b = *pb;

  if (a < b) { pR->inf = a; pR->sup = b; }
  else       { pR->inf = b; pR->sup = a; }
} /* BiasKJoinRR */

VOID BiasKJoinRI (BIASKINTERVAL * const pR,
		 const REAL * const pa, const BIASKINTERVAL * const pB)
// BiasK
{
  REAL a = *pa;

  pR->inf = (a < pB->inf) ? a : pB->inf;
  pR->sup = (a > pB->sup) ? a : pB->sup;
} /* BiasKJoinRI */

VOID BiasKJoinII (BIASKINTERVAL * const pR,
		 const BIASKINTERVAL * const pA, const BIASKINTERVAL * const pB)
// BiasK
{
  pR->inf = (pA->inf < pB->inf) ? pA->inf : pB->inf;
  pR->sup = (pA->sup > pB->sup) ? pA->sup : pB->sup;
} /* BiasKJoinII */

// ****************************************
// BiasK: for compatibility with old version
INT     BiasKIntersection(BIASKINTERVAL       * const pR, const BIASKINTERVAL * const pA, const BIASKINTERVAL * const pB)
	{BiasKMeetII(pR,pA,pB);return BiasKIsProper(pR);}
VOID    BiasKHullR       (BIASKINTERVAL       * const pR, const REAL         * const pa)
	{BiasKJoinRR(pR,pa,pa);}
VOID    BiasKHullRR      (BIASKINTERVAL       * const pR, const REAL         * const pa, const REAL         * const pb)
	{BiasKJoinRR(pR,pa,pb);}
VOID    BiasKHullRI      (BIASKINTERVAL       * const pR, const REAL         * const pa, const BIASKINTERVAL * const pB)
	{BiasKJoinRI(pR,pa,pB);}
VOID    BiasKHullII      (BIASKINTERVAL       * const pR, const BIASKINTERVAL * const pA, const BIASKINTERVAL * const pB)
	{BiasKJoinII(pR,pA,pB);}
// ****************************************

INT BiasKInR (const REAL * const pa, const BIASKINTERVAL * const pB)
// BiasK
  /**********************************************************************
  *  Result: 1 if a is contained in B, 0 otherwise
  */
{
  REAL a = *pa;

  return ((a >= pB->inf) && (a <= pB->sup));
} /* BiasKInR */

INT BiasKInI (const BIASKINTERVAL * const pA, const BIASKINTERVAL * const pB)
// BiasK
  /**********************************************************************
  *  Result: 1 if A is contained in B, 0 otherwise
  */
{
  return ((pA->inf >= pB->inf) && (pA->sup <= pB->sup));
} /* BiasKInI */

INT BiasKInInteriorR (const REAL * const pa, const BIASKINTERVAL * const pB)
// BiasK
  /**********************************************************************
  *  Result: 1 if a is contained in the interior of B, 0 otherwise
  */
{
  REAL a = *pa;

  return ((a > pB->inf) && (a < pB->sup));
} /* BiasKInInteriorR */

INT BiasKInInteriorI (const BIASKINTERVAL * const pA, const BIASKINTERVAL * const pB)
// BiasK
  /**********************************************************************
  *  Result: 1 if A is contained in the interior of B, 0 otherwise
  */
{
  return ((pA->inf > pB->inf) && (pA->sup < pB->sup));
} /* BiasKInInteriorI */

INT BiasKIsEqual (const BIASKINTERVAL * const pA, const BIASKINTERVAL * const pB)
// BiasK
  /**********************************************************************
  *  Result: 1, if A == B, 0 otherwise
  */
{
  return ((pA->inf == pB->inf) && (pA->sup == pB->sup));
} /* BiasKIsEqual */

INT BiasKIsProper (const BIASKINTERVAL * const pA)
// BiasK
{
  return (pA->inf <= pA->sup);
} /* BiasKIsProper */
INT BiasKIsImproper (const BIASKINTERVAL * const pA)
// BiasK
{
  return (pA->inf >= pA->sup);
} /* BiasKIsImproper */
