/************************************************************************
 *
 * Basic Interval Arithmetic Subroutines Level 1
 * Modified for generalized intervals and Kaucher arithmetic (2005/06 A. Goldsztejn)
* ---------------------------------------------
 *
 * Copyright (c) 1993, 1997 Olaf Knueppel
 *
 * $Id: BiasK1.c,v 1.3 2000/02/08 08:40:13 knueppel Exp $
 *
 ************************************************************************/

#include "BiasK1.h"
#include "BiasRnd.h"
#include "BiasKInt.h"
#include <string.h>

#if !defined (__BIASKINFSUP__) || !defined (__BIASKIEEE__)
#error __BIASKINFSUP__ and __BIASKIEEE__ not set
#endif

static const char ErrorVectorDivideByZero[] = "Vector Division by Zero";

/************************************************************************
 *  Implementation of global functions
 ************************************************************************/

VOID BiasKPredVR (REAL *pr,
                 const REAL *pa,
                 const int dim)
  /**********************************************************************
  *  r_i = largest possible floating point number with r_i < a_i
  */
// BiasKK
{
  INT i;
  REAL Eta = BiasKEta;

  _BiasKRoundDown ();
  for (i = 0; i < dim; i++) *(pr++) = *(pa++) - Eta;
  _SetRoundToNearest();
}

INT BiasKPredVI (BIASKINTERVAL * const pR_start,
                const BIASKINTERVAL * const pA_start,
		const int dim)
  /**********************************************************************
  *  R_i = largest possible interval with R_i.inf > A_i.inf and R_i.sup < A_i.sup
  *  Result: 1, if R is a true interval vector, 0 otherwise
  */
// BiasKK
{
  INT i;
  INT ok = 1;
  BIASKINTERVAL *pR;
  const BIASKINTERVAL *pA;
  REAL Eta = BiasKEta;

  _BiasKRoundDown ();
  pR = pR_start; pA = pA_start;
  for (i = 0; i < dim; i++) (pR++)->sup = (pA++)->sup - Eta;
  _BiasKRoundUp ();
  pR = pR_start; pA = pA_start;
  for (i = 0; i < dim; i++) (pR++)->inf = (pA++)->inf + Eta;
  pR = pR_start;
  for (i = 0; i < dim; i++, pR++) ok &= (pR->inf <= pR->sup);
  _SetRoundToNearest();
  return ok;
}

VOID BiasKSuccVR (REAL *pr,
                 const REAL *pa,
                 const int dim)
  /**********************************************************************
  *  r_i = smallest possible floating point number with r_i > a_i
  */
// BiasKK
{
  INT i;
  REAL Eta = BiasKEta;

  _BiasKRoundUp ();
  for (i = 0; i < dim; i++) *(pr++) = *(pa++) + Eta;
  _SetRoundToNearest();
}

VOID BiasKSuccVR2VI (BIASKINTERVAL * const pR_start,
                 const REAL * const pa_start,
		 const int dim)
{
  INT i;
  BIASKINTERVAL *pR;
  const REAL *pa;
  REAL Eta = BiasKEta;

  _BiasKRoundDown ();
  pR = pR_start; pa = pa_start;
  for (i = 0; i < dim; i++) (pR++)->inf = *(pa++) - Eta;
  _BiasKRoundUp ();
  pR = pR_start; pa = pa_start;
  for (i = 0; i < dim; i++) (pR++)->sup = *(pa++) + Eta;
}

VOID BiasKSuccVI (BIASKINTERVAL * const pR_start,
                 const BIASKINTERVAL * const pA_start,
		 const int dim)
  /**********************************************************************
  *  R_i = smallest possible interval with R_i.inf < A_i.inf
  *        and R_i.sup > A_i.sup
  */
// BiasKK
{
  INT i;
  BIASKINTERVAL *pR;
  const BIASKINTERVAL *pA;
  REAL Eta = BiasKEta;

  _BiasKRoundDown ();
  pR = pR_start; pA = pA_start;
  for (i = 0; i < dim; i++) (pR++)->inf = (pA++)->inf - Eta;
  _BiasKRoundUp ();
  pR = pR_start; pA = pA_start;
  for (i = 0; i < dim; i++) (pR++)->sup = (pA++)->sup + Eta;
}

VOID BiasKAddVRVR (BIASKINTERVAL * const pR_start,
                  const REAL * const pa_start,
		  const REAL * const pb_start,
                  const int dim)
  /**********************************************************************
   *  vR = va + vb
   */
// BiasKK
{
  INT i;
  const REAL *pa, *pb;
  BIASKINTERVAL *pR;

  _BiasKRoundDown ();
  pa = pa_start; pb = pb_start; pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->inf = *(pa++) + *(pb++);
  _BiasKRoundUp ();
  pa = pa_start; pb = pb_start; pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->sup = *(pa++) + *(pb++);
  _SetRoundToNearest();
}

VOID BiasKAddVRVI (BIASKINTERVAL * const pR_start,
                  const REAL * const pa_start,
		  const BIASKINTERVAL * const pB_start,
                  const int dim)
  /**********************************************************************
   *  vR = va + vB
   */
// BiasKK
{
  INT i;
  const REAL *pa;
  const BIASKINTERVAL *pB;
  BIASKINTERVAL *pR;

  _BiasKRoundDown ();
  pa = pa_start; pB = pB_start;
  pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->inf = *(pa++) + (pB++)->inf;
  _BiasKRoundUp ();
  pa = pa_start; pB = pB_start;
  pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->sup = *(pa++) + (pB++)->sup;
  _SetRoundToNearest();
}

VOID BiasKAddVIVI (BIASKINTERVAL * const pR_start,
                  const BIASKINTERVAL * const pA_start,
		  const BIASKINTERVAL * const pB_start,
                  const int dim)
  /**********************************************************************
   *  vR = vA + vB
   */
// BiasKK
{
  INT i;
  const BIASKINTERVAL *pA, *pB;
  BIASKINTERVAL *pR;

  _BiasKRoundDown ();
  pA = pA_start; pB = pB_start;
  pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->inf = (pA++)->inf + (pB++)->inf;//
//  _BiasKRoundUp ();
  pA = pA_start; pB = pB_start;
  pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->sup = (pA++)->sup + (pB++)->sup;
//  _SetRoundToNearest();
}

VOID BiasKSubVRVR (BIASKINTERVAL * const pR_start,
                  const REAL * const pa_start,
		  const REAL * const pb_start,
                  const int dim)
  /**********************************************************************
   *  vR = va - vb
   */
// BiasKK
{
  INT i;
  const REAL *pa, *pb;
  BIASKINTERVAL *pR;

  _BiasKRoundDown ();
  pa = pa_start; pb = pb_start; pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->inf = *(pa++) - *(pb++);
  _BiasKRoundUp ();
  pa = pa_start; pb = pb_start; pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->sup = *(pa++) - *(pb++);
  _SetRoundToNearest();
}

VOID BiasKSubVRVI (BIASKINTERVAL * const pR_start,
                  const REAL * const pa_start,
		  const BIASKINTERVAL * const pB_start,
                  const int dim)
  /**********************************************************************
   *  vR = va - vB
   *  Assertion: vR != vB
   */
// BiasKK
{
  INT i;
  const REAL *pa;
  const BIASKINTERVAL *pB;
  BIASKINTERVAL *pR;

  _BiasKRoundDown ();
  pa = pa_start; pB = pB_start; pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->inf = *(pa++) - (pB++)->sup;
  _BiasKRoundUp ();
  pa = pa_start; pB = pB_start; pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->sup = *(pa++) - (pB++)->inf;
  _SetRoundToNearest();
}

VOID BiasKSubVIVR (BIASKINTERVAL * const pR_start,
                  const BIASKINTERVAL * const pA_start,
		  const REAL * const pb_start,
                  const int dim)
  /**********************************************************************
   *  vR = vA - vb
   */
// BiasKK
{
  INT i;
  const REAL *pb;
  const BIASKINTERVAL *pA;
  BIASKINTERVAL *pR;

  _BiasKRoundDown ();
  pA = pA_start; pb = pb_start; pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->inf = (pA++)->inf - *(pb++);
  _BiasKRoundUp ();
  pA = pA_start; pb = pb_start; pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->sup = (pA++)->sup - *(pb++);
  _SetRoundToNearest();
}

VOID BiasKSubVIVI (BIASKINTERVAL * const pR_start,
                  const BIASKINTERVAL * const pA_start,
		  const BIASKINTERVAL * const pB_start,
                  const int dim)
  /**********************************************************************
   *  vR = vA - vB
   *  Assertion: vR != vB
   */
// BiasKK
{
  INT i;
  const BIASKINTERVAL *pA, *pB;
  BIASKINTERVAL *pR;

  _BiasKRoundDown ();
  pA = pA_start; pB = pB_start; pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->inf = (pA++)->inf - (pB++)->sup;
  _BiasKRoundUp ();
  pA = pA_start; pB = pB_start; pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->sup = (pA++)->sup - (pB++)->inf;
  _SetRoundToNearest();
}

VOID BiasKMulRVR (BIASKINTERVAL * const pR_start,
                 const REAL *pa,
		 const REAL * const pb_start,
                 const int dim)
  /**********************************************************************
   *  vR = a * vb
   */
// BiasKK
{
  INT i;
  REAL a = *pa;
  const REAL *pb;
  BIASKINTERVAL *pR;

  _BiasKRoundDown ();
  pb = pb_start; pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->inf = a * *(pb++);
  _BiasKRoundUp ();
  pb = pb_start; pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->sup = a * *(pb++);
  _SetRoundToNearest();
}

VOID BiasKMulIVR (BIASKINTERVAL * const pR_start,
                 const BIASKINTERVAL * const pA,
		 const REAL * const pb_start,
                 const int dim)
  /**********************************************************************
   *  vR = A * vb
   */
// BiasKK
{
  INT i;
  REAL painf = pA->inf;
  REAL pasup = pA->sup;
  const REAL *pb;
  BIASKINTERVAL *pR;

  _BiasKRoundDown ();
  pb = pb_start; pR = pR_start;
  for (i = 0; i < dim; i++) {
    (pR++)->inf = *pb * ((Sign (*pb) == 0) ? painf : pasup);
    pb++;
  }
  _BiasKRoundUp ();
  pb = pb_start; pR = pR_start;
  for (i = 0; i < dim; i++) {
    (pR++)->sup = *pb * ((Sign (*pb) == 0) ? pasup : painf);
    pb++;
  }
  _SetRoundToNearest();
}

VOID BiasKMulRVI (BIASKINTERVAL * const pR_start,
                 const REAL * const pa,
		 const BIASKINTERVAL * const pB_start,
                 const int dim)
  /**********************************************************************
   *  vR = a * vB
   *  Assertion: vR != vB
   */
// BiasKK
{
  INT i;
  const REAL a = *pa;
  const BIASKINTERVAL *pB;
  BIASKINTERVAL *pR;

  _BiasKRoundDown ();
  pB = pB_start; pR = pR_start;
  if (Sign (*pa) == 0) {                               /* a >= 0 */
    for (i = 0; i < dim; i++) (pR++)->inf = a * (pB++)->inf;
    _BiasKRoundUp ();
    pB = pB_start; pR = pR_start;
    for (i = 0; i < dim; i++) (pR++)->sup = a * (pB++)->sup;
  }
  else {                                               /* a <  0 */
    for (i = 0; i < dim; i++) (pR++)->inf = a * (pB++)->sup;
    _BiasKRoundUp ();
    pB = pB_start; pR = pR_start;
    for (i = 0; i < dim; i++) (pR++)->sup = a * (pB++)->inf;
  }
  _SetRoundToNearest();
}

VOID BiasKMulIVI (BIASKINTERVAL *pR,
                 const BIASKINTERVAL * const pA,
		 const BIASKINTERVAL *pB,
                 const int dim)
  /**********************************************************************
   *  vR = A * vB
   *  Assertion: vR != vB
   */
// BiasKK ??
{
  INT i;

  for (i=0; i<dim; i++)
  {
      BiasKMulII(pR++,pA,pB++);
  }    
}

VOID BiasKMacRVR (BIASKINTERVAL * const pR_start,
                 const REAL * const pa,
		 const REAL * const pb_start,
                 const int dim)
  /**********************************************************************
   *  vR += a * vb
   */
// BiasKK
{
  INT i;
  const REAL a = *pa;
  const REAL * pb;
  BIASKINTERVAL *pR;

  _BiasKRoundDown ();
  pb = pb_start; pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->inf += a * *(pb++);
  _BiasKRoundUp ();
  pb = pb_start; pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->sup += a * *(pb++);
  _SetRoundToNearest();
}

VOID BiasKMacIVR (BIASKINTERVAL * const pR_start,
                 const BIASKINTERVAL * const pA,
		 const REAL * const pb_start,
                 const int dim)
  /**********************************************************************
   *  vR += A * vb
   */
// BiasKK
{
  INT i;
  REAL painf = pA->inf;
  REAL pasup = pA->sup;
  const REAL *pb;
  BIASKINTERVAL *pR;

  _BiasKRoundDown ();
  pb = pb_start; pR = pR_start;
  for (i = 0; i < dim; i++) {
    (pR++)->inf += *pb * ((Sign (*pb) == 0) ? painf : pasup);
    pb++;
  }
  _BiasKRoundUp ();
  pb = pb_start; pR = pR_start;
  for (i = 0; i < dim; i++) {
    (pR++)->sup += *pb * ((Sign (*pb) == 0) ? pasup : painf);
    pb++;
  }
  _SetRoundToNearest();
}

VOID BiasKMacRVI (BIASKINTERVAL * const pR_start,
                 const REAL * const pa,
		 const BIASKINTERVAL * const pB_start,
                 const int dim)
  /**********************************************************************
   *  vR += a * vB
   *  Assertion: vR != vB
   */
// BiasKK
{
  INT i;
  const REAL a = *pa;
  const BIASKINTERVAL *pB;
  BIASKINTERVAL *pR;

  _BiasKRoundDown ();
  pB = pB_start; pR = pR_start;
  if (Sign (*pa) == 0) {                               /* a >= 0 */
    for (i = 0; i < dim; i++) (pR++)->inf += a * (pB++)->inf;
    _BiasKRoundUp ();
    pB = pB_start; pR = pR_start;
    for (i = 0; i < dim; i++) (pR++)->sup += a * (pB++)->sup;
  }
  else {                                               /* a <  0 */
    for (i = 0; i < dim; i++) (pR++)->inf += a * (pB++)->sup;
    _BiasKRoundUp ();
    pB = pB_start; pR = pR_start;
    for (i = 0; i < dim; i++) (pR++)->sup += a * (pB++)->inf;
  }
  _SetRoundToNearest();
}

VOID BiasKMacIVI (BIASKINTERVAL *pR,
                 const BIASKINTERVAL * const pA,
		 const BIASKINTERVAL *pB,
                 const int dim)
  /**********************************************************************
   *  vR += A * vB
   */
// BiasKK ??
{
  INT i;
  
  for(i=0; i<dim; i++) BiasKMacII(pR++,pA,pB++);
}

VOID BiasKMacRVIs (BIASKINTERVAL * const pR_start,
                  const REAL * const pa,
		  const BIASKINTERVAL * const pB_start,
                  const int dim,
		  const int Bstep)
  /**********************************************************************
   *  vR += a * vB, Stepwidth in B = Bstep
   *  Assertion: vR != vB
   */
// BiasKK
{
  INT i;
  const REAL a = *pa;
  const BIASKINTERVAL *pB;
  BIASKINTERVAL *pR;

  _BiasKRoundDown ();
  pB = pB_start; pR = pR_start;
  if (Sign (*pa) == 0) {                               /* a >= 0 */
    for (i = 0; i < dim; i++, pB += Bstep)
      (pR++)->inf += a * pB->inf;
    _BiasKRoundUp ();
    pB = pB_start; pR = pR_start;
    for (i = 0; i < dim; i++, pB += Bstep)
      (pR++)->sup += a * pB->sup;
  }
  else {                                               /* a <  0 */
    for (i = 0; i < dim; i++, pB += Bstep)
      (pR++)->inf += a * pB->sup;
    _BiasKRoundUp ();
    pB = pB_start; pR = pR_start;
    for (i = 0; i < dim; i++, pB += Bstep)
      (pR++)->sup += a * pB->inf;
  }
  _SetRoundToNearest();
}

VOID BiasKMacsRVIs (BIASKINTERVAL * const pR_start,
                   const REAL * const pa,
		   const BIASKINTERVAL * const pB_start,
                   const int dim,
		   const int Rstep,
                   const int Bstep)
  /**********************************************************************
   *  vR += a * vB, Stepwidth in B = Bstep, Stepwidth in R = Rstep
   *  Assertion: vR != vB
   */
// BiasKK
{
  INT i;
  const REAL a = *pa;
  const BIASKINTERVAL *pB;
  BIASKINTERVAL *pR;

  _BiasKRoundDown ();
  pB = pB_start; pR = pR_start;
  if (Sign (*pa) == 0) {                               /* a >= 0 */
    for (i = 0; i < dim; i++, pB += Bstep, pR += Rstep)
      pR->inf += a * pB->inf;
    _BiasKRoundUp ();
    pB = pB_start; pR = pR_start;
    for (i = 0; i < dim; i++, pB += Bstep, pR += Rstep)
      pR->sup += a * pB->sup;
  }
  else {                                               /* a <  0 */
    for (i = 0; i < dim; i++, pB += Bstep, pR += Rstep)
      pR->inf += a * pB->sup;
    _BiasKRoundUp ();
    pB = pB_start; pR = pR_start;
    for (i = 0; i < dim; i++, pB += Bstep, pR += Rstep)
      pR->sup += a * pB->inf;
  }
  _SetRoundToNearest();
}

VOID BiasKDivVRR (BIASKINTERVAL * const pR_start,
                 const REAL * const pa_start,
		 const REAL * const pb,
                 const int dim)
  /**********************************************************************
   *  vR = va / b
   */
// BiasKK
{
  INT i;
  const REAL b = *pb;
  const REAL *pa;
  BIASKINTERVAL *pR;

#if defined (__BIASKRAISEDIVIDEBYZERO__)
  if (Zero (*pb)) _BiasKError (ErrorVectorDivideByZero);
#endif
  _BiasKRoundDown ();
  pa = pa_start; pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->inf = *(pa++) / b;
  _BiasKRoundUp ();
  pa = pa_start; pR = pR_start;
  for (i = 0; i < dim; i++) (pR++)->sup = *(pa++) / b;
  _SetRoundToNearest();
}

VOID BiasKDivVIR (BIASKINTERVAL * pR,
                 const BIASKINTERVAL * pA,
		 const REAL * const pb,
                 const int dim)
  /**********************************************************************
   *  vR = vA / b
   *  Assertion: vR != vA
   */
// BiasKK ( MODIFIED )
{
  INT i;

  for (i = 0; i < dim; i++) BiasKDivIR(pR++, pA++, pb);
}

VOID BiasKDivVRI (BIASKINTERVAL * pR,
                 const REAL * pa,
		 const BIASKINTERVAL *const pB,
                 const int dim)
  /**********************************************************************
   *  vR = va / B
   *  Assertion: vR != B
   */
// BiasKK ( MODIFIED )
{
  INT i;

  for (i = 0; i < dim; i++) BiasKDivRI(pR++, pa++, pB);
}

VOID BiasKDivVII (BIASKINTERVAL *pR,
                 const BIASKINTERVAL *pA,
		 const BIASKINTERVAL * const pB,
                 const int dim)
  /**********************************************************************
   *  vR = vA / B
   *  Assertion: vR != vA and vR != B
   */
// BiasKK
{
  INT i;

  for (i = 0; i < dim; i++) BiasKDivII(pR++, pA++, pB);
}

VOID BiasKMacVRVR(BIASKINTERVAL *const pR,
                 const REAL * const pa_start,
		 const REAL * const pb_start,
                 const int dim)
  /**********************************************************************
   *  R = va * vb
   */
// BiasKK
{
  INT i;
  const REAL *pa, *pb;
  REAL sum;

  _BiasKRoundDown ();
  pa = pa_start; pb = pb_start; sum = 0.0;
  for (i = 0; i < dim; i++) sum += *(pa++) * *(pb++);
  pR->inf = sum;
  _BiasKRoundUp ();
  pa = pa_start; pb = pb_start; sum = 0.0;
  for (i = 0; i < dim; i++) sum += *(pa++) * *(pb++);
  pR->sup = sum;
  _SetRoundToNearest();
}

VOID BiasKMacVRVI(BIASKINTERVAL *pR,
                 const REAL * pa,
		 const BIASKINTERVAL * pB,
                 const int dim)
  /**********************************************************************
   *  R = va * vB
   */
// BiasKK ( MODIFIED )
{
  INT i;

  pR->sup = pR->inf = 0.0;
  for (i = 0; i < dim; i++) BiasKMacRI(pR, pa++, pB++);
}

VOID BiasKMacVIVI(BIASKINTERVAL * const pR,
                 const BIASKINTERVAL *pA,
		 const BIASKINTERVAL *pB,
                 const int dim)
  /**********************************************************************
   *  R = vA * vB
   */
// BiasKK
{
  INT i;

  pR->sup = pR->inf = 0.0;
  for (i = 0; i < dim; i++) BiasKMacII(pR, pA++, pB++);
}

VOID BiasKInfV (REAL *pr,
               const BIASKINTERVAL *pA,
               const int dim)
  /**********************************************************************
  *  vr = Inf (vA)
  */
// BiasKK
{
  INT i;

  for (i = 0; i < dim; i++) *(pr++) = (pA++)->inf;
}

VOID BiasKSupV (REAL *pr,
               const BIASKINTERVAL *pA,
               const int dim)
  /**********************************************************************
  *  vr = Sup (vA)
  */
// BiasKK
{
  INT i;

  for (i = 0; i < dim; i++) *(pr++) = (pA++)->sup;
}

VOID BiasKNegV (BIASKINTERVAL *pR,
               const BIASKINTERVAL *pA,
               const int dim)
  /**********************************************************************
  *  vR = - vA
  */
// BiasKK
{
    INT i;

  for ( i = 0; i < dim; i++, pR++, pA++) BiasKNeg(pR,pA);
}
// **************************************************************************
VOID BiasKDualV (BIASKINTERVAL * pR, const BIASKINTERVAL * pA, const INT dim)
// BiasKK
{
    INT i;
  for ( i = 0; i < dim; i++, pR++, pA++) BiasKDual(pR,pA);
}
VOID BiasKProV (BIASKINTERVAL * pR, const BIASKINTERVAL * pA, const INT dim)
// BiasKK
{
    INT i;

  for ( i = 0; i < dim; i++, pR++, pA++) BiasKPro(pR,pA);
}
VOID BiasKImpV (BIASKINTERVAL * pR, const BIASKINTERVAL * pA, const INT dim)
// BiasKK
{
    INT i;

  for ( i = 0; i < dim; i++, pR++, pA++) BiasKImp(pR,pA);
}
// **************************************************************************

VOID BiasKMidV (REAL *pr,
               const BIASKINTERVAL *pA,
               const int dim)
  /**********************************************************************
  *  vr = Midpoint (vA)
  */
// BiasKK ( MODIFIED )
{
  INT i;

  _BiasKRoundUp ();
  for (i = 0; i < dim; i++, pA++)
    *(pr++) = BiasKMid(pA) ;
  _SetRoundToNearest();
}

VOID BiasKMidRadV (REAL *pm,
                  REAL *pr,
                  const BIASKINTERVAL *pA,
                  const int dim)
  /**********************************************************************
  *  vm = Midpoint (vA) and vr = Radius (vA)
  */
// BiasKK
{
  INT i;
  REAL r;

  for (i = 0; i < dim; i++, pA++) { BiasKMidRad(pm++,pr++,pA); }
}

VOID BiasKDiamV (REAL *pr,
                const BIASKINTERVAL *pA,
                const int dim)
  /**********************************************************************
  *  vr = Diameter (vA)
  */
// BiasKK ( MODIFIED )
{
  INT i;

  for (i = 0; i < dim; i++, pA++) *(pr++) = BiasKDiam(pA);
}

VOID BiasKAbsV (REAL *pr,
               const BIASKINTERVAL *pA,
               const int dim)
  /**********************************************************************
  *  vr = Abs (vA)
  */
// BiasKK
{
  INT i;

  for (i = 0; i < dim; i++) *(pr++) = BiasKAbs (pA++);
}

VOID    BiasKVRVR2VI     (BIASKINTERVAL *pR,
                         const REAL *pa,
                         const REAL *pb,
                         const INT dim)
{
  INT i;

  for (i = 0; i < dim; i++)
    BiasKRR2I (pR++, pa++, pb++);
}

VOID BiasKMeetVRVR (BIASKINTERVAL *pR,
                       const REAL *pa,
		       const REAL *pb,
                       const int dim)
// BiasKK ( MODIFIED )
{
  INT i;

  for (i = 0; i < dim; i++)
    BiasKMeetRR (pR++, pa++, pb++);
}
VOID BiasKMeetVRVI (BIASKINTERVAL *pR,
                       const REAL *pa,
		       const BIASKINTERVAL *pB,
                       const int dim)
// BiasKK ( MODIFIED )
{
  INT i;

  for (i = 0; i < dim; i++)
    BiasKMeetRI (pR++, pa++, pB++);
}
VOID BiasKMeetVIVI (BIASKINTERVAL *pR,
                       const BIASKINTERVAL *pA,
		       const BIASKINTERVAL *pB,
                       const int dim)
// BiasKK ( MODIFIED )
{
  INT i;

  for (i = 0; i < dim; i++)
    BiasKMeetII (pR++, pA++, pB++);
}

VOID BiasKJoinVRVR (BIASKINTERVAL *pR,
                   const REAL *pa,
                   const REAL *pb,
                   const int dim)
// BiasKK
{
  INT i;

  for (i = 0; i < dim; i++) BiasKJoinRR (pR++, pa++, pb++);
}

VOID BiasKJoinVRVI (BIASKINTERVAL *pR,
                   const REAL *pa,
		   const BIASKINTERVAL *pB,
                   const int dim)
// BiasKK
{
  INT i;

  for (i = 0; i < dim; i++) BiasKJoinRI (pR++, pa++, pB++);
}

VOID BiasKJoinVIVI (BIASKINTERVAL *pR,
                   const BIASKINTERVAL *pA,
		   const BIASKINTERVAL *pB,
                   const int dim)
// BiasKK
{
  INT i;

  for (i = 0; i < dim; i++) BiasKJoinII (pR++, pA++, pB++);
}

INT BiasKInVR (const REAL *pa,
              const BIASKINTERVAL *pB,
              const int dim)
  /**********************************************************************
  *  Result: 1 if va is contained in vB, 0 otherwise
  */
// BiasKK
{
  INT i;

  for (i = 0; i < dim; i++)
    if (!BiasKInR (pa++, pB++)) return 0;
  return 1;
}

INT BiasKInVI (const BIASKINTERVAL *pA,
              const BIASKINTERVAL *pB,
              const int dim)
  /**********************************************************************
  *  Result: 1 if vA is contained in vB, 0 otherwise
  */
// BiasKK
{
  INT i;

  for (i = 0; i < dim; i++)
    if (!BiasKInI (pA++, pB++)) return 0;
  return 1;
}

INT BiasKInInteriorVR (const REAL *pa,
                      const BIASKINTERVAL *pB,
                      const int dim)
  /**********************************************************************
  *  Result: 1 if va is contained in the interior of vB, 0 otherwise
  */
// BiasKK
{
  INT i;

  for (i = 0; i < dim; i++)
    if (!BiasKInInteriorR (pa++, pB++)) return 0;
  return 1;
}

INT BiasKInInteriorVI (const BIASKINTERVAL *pA,
                      const BIASKINTERVAL *pB,
                      const int dim)
  /**********************************************************************
  *  Result: 1 if vA is contained in the interior of vB, 0 otherwise
  */
// BiasKK
{
  INT i;

  for (i = 0; i < dim; i++)
    if (!BiasKInInteriorI (pA++, pB++)) return 0;
  return 1;
}

INT BiasKIsEqualV (const BIASKINTERVAL *pA,
                  const BIASKINTERVAL *pB,
                  const int dim)
  /**********************************************************************
  *  Result: 1, if vA == vB, 0 otherwise
  */
// BiasKK
{
  INT i;

  for (i = 0; i < dim; i++)
    if (!BiasKIsEqual (pA++, pB++)) return 0;
  return 1;
}

VOID BiasKSetToZeroV (BIASKINTERVAL * const pR,
                     const int dim)
  /**********************************************************************
  *  vR = 0
  */
// BiasKK
{
  memset (pR, 0, dim * sizeof (BIASKINTERVAL));
}

VOID BiasKSetToZeroVs (BIASKINTERVAL *pR,
                      const int dim,
		      const int step)
  /**********************************************************************
  *  vR = 0, Stepwidth of R = step
  */
// BiasKK
{
  INT j;

    for (j = 0; j < dim; j++, pR += step) pR->sup = pR->inf = 0.0;
}

// BiasK: new fonctions
// ****************************************
INT BiasKIsProperV (const BIASKINTERVAL * pA, int dim)
{
  INT i,test;

  for(i=0,test=1; i<dim & test; i++,pA++)
    test=test & BiasKIsProper(pA);

  return test;
}

INT BiasKIsImproperV (const BIASKINTERVAL * pA, int dim)
{
  INT i,test;

  for(i=0,test=1; i<dim & test ; i++,pA++)
    test=test & BiasKIsImproper(pA);

  return test;
}

// ****************************************
