/************************************************************************
 *
 * Basic Interval Arithmetic Subroutines Standard Functions
 * Modified for generalized intervals and Kaucher arithmetic (2005/06 A. Goldsztejn)
 * --------------------------------------------------------
 *
 * Copyright (c) 1993, 1995, 1997 Olaf Knueppel
 *
 * $Id: BiasKF.c,v 1.1.1.1 2004/04/21 10:02:36 dalmas Exp $
 *
 ************************************************************************/

#include "BiasKF.h"
#include "BiasRnd.h"
#include "BiasKInt.h"
#include <math.h>

REAL BiasKPi, BiasKTwoPi, BiasKPiHalf, BiasKPiQuarter;
REAL BiasKE, BiasKSqrt2, BiasKInvSqrt2, BiasKLn10;

static BIASKINTERVAL BiasKPiHalfIncl; /* enclosure of Pi/2  */
static BIASKINTERVAL BiasKLn10Incl;   /* enclosure of Ln 10 */

static BIASKINTERVAL PlusMinusOne;   /* [-1,1] */

static REAL BiasKStdFuncOnePlusEps;  /* (1.0 + n * Epsilon)          */
static REAL BiasKStdFuncOneMinusEps; /* (1.0 - n * Epsilon)          */

static REAL Zero     =  0.0;
static REAL MinusOne = -1.0;
static REAL PlusOne  =  1.0;
static REAL PlusTwo  =  2.0;

#define RoundUp(x) \
  (((x < 0.0) ? BiasKStdFuncOneMinusEps : BiasKStdFuncOnePlusEps) * (x) \
   + BiasKEta)
#define RoundDown(x) \
  (((x < 0.0) ? BiasKStdFuncOnePlusEps : BiasKStdFuncOneMinusEps) * (x) \
   - BiasKEta)
#define DEBUGGING

/************************************************************************
*  Implementation of local functions
*************************************************************************/

static INT Quadrant (const REAL x)
  /**********************************************************************
   *  Calculation of the quadrant x is lying in (x in [0, 2Pi])
   */
{
  if (x <= BiasKPiHalf) return 0;
  if (x <= BiasKPi) return 1;
  if (x <= BiasKPi + BiasKPiHalf) return 2;
  return 3;
}

static VOID ScaleTo2Pi (REAL * const py_inf,
                        REAL * const py_sup,
                        const BIASKINTERVAL * const pX)
  /**********************************************************************
   *  The interval x is scaled to the range [0, 2Pi].
   *  Result: Interval [y_inf, y_sup]
   */
{
  REAL q;
  REAL x_inf, x_sup;

  if (((*py_inf = BiasKInf (pX)) >= 0.0)
      && ((*py_sup = BiasKSup (pX)) < BiasKPiHalf)) return;

  q = BiasKInf (pX) / BiasKTwoPi;
  q = floor(BiasKPredR (& q));
  q *= BiasKTwoPi;
  q = BiasKSuccR (& q);
  if (q > 0.0) {
    q *= (1.0 + BiasKEpsilon); /* error estimation: 1 unit in LSB */
    q = BiasKSuccR (& q);
  }
  x_inf = BiasKInf (pX) - q;
  x_inf = BiasKPredR (& x_inf);
  if (x_inf < 0.0) {
    x_inf += BiasKTwoPi;
    x_inf = BiasKPredR (& x_inf);
  }
  q = BiasKSup (pX) / BiasKTwoPi;
  q = floor(BiasKPredR (& q));
  q *= BiasKTwoPi;
  q = BiasKPredR (& q);
  if (q < 0.0) {
    q *= (1.0 + BiasKEpsilon); /* error estimation: 1 unit in LSB */
    q = BiasKPredR (& q);
  }
  x_sup = BiasKSup (pX) - q;
  x_sup = BiasKSuccR (& x_sup);
  if (x_sup > BiasKTwoPi) {
    x_sup -= BiasKTwoPi;
    x_sup = BiasKSuccR (& x_sup);
  }
#ifdef DEBUGGING
  if (x_inf < 0.0)
    _BiasKError ("ScaleTo2Pi (): inf < 0");
  if (x_sup > BiasKTwoPi)
    _BiasKError ("ScaleTo2Pi (): sup > 2Pi");
#endif
  *py_inf = x_inf; *py_sup = x_sup;
}

/************************************************************************
*  Implementation of global functions
*************************************************************************/

VOID BiasKFuncInit (VOID)
  /**********************************************************************
   *  Initialization
   */
{
  INT InvalidBits = __BIASSTDFUNCINVALIDBITS__;
  BIASKINTERVAL tmp;

  BiasKInit ();

  BiasKPi         = 3.14159265358979323846;
  BiasKTwoPi      = 6.28318530717958647693;
  BiasKPiHalf     = 1.57079632679489661923;
  BiasKPiQuarter  = 0.78539816339744830962;
  BiasKE          = 2.7182818284590452354;
  BiasKSqrt2      = 1.41421356237309504880;
  BiasKInvSqrt2   = 0.70710678118654752440;
  BiasKLn10       = 2.30258509299404568402;

  BiasKSuccR2I(& BiasKPiHalfIncl, & BiasKPiHalf);

  BiasKSuccR2I(& BiasKLn10Incl, & BiasKLn10);

  BiasKJoinRR (& PlusMinusOne, & MinusOne, & PlusOne);

  BiasKRoundNear ();

  BiasKStdFuncOnePlusEps  = 1.0 + (1 << (InvalidBits + 1)) * BiasKEpsilon;
  BiasKStdFuncOneMinusEps = 1.0 - (1 << (InvalidBits + 1)) * BiasKEpsilon;
}

VOID BiasKSinProper(BIASKINTERVAL * const pR,
              const BIASKINTERVAL * const pX)
  /**********************************************************************
   *  R = Sin (X)
   */
{
#define Maximum(a,b) (((a) > (b)) ? (a) : (b))
#define Minimum(a,b) (((a) < (b)) ? (a) : (b))
  REAL x_inf, x_sup;
  REAL y_inf, y_sup;
  int q_inf, q_sup;

  if (BiasKDiam (pX) >= BiasKTwoPi) {
    BiasKJoinRR (pR, & MinusOne, & PlusOne);
    return;
  }
  /* Reduction of x to [0, 2Pi] */
  ScaleTo2Pi (& x_inf, & x_sup, pX);
  if (x_sup >= x_inf + BiasKTwoPi) { /* security */
    BiasKJoinRR (pR, & MinusOne, & PlusOne);
    return;
  }
  /* x_inf, x_sup are now in the range [0, 2Pi]
   * Quadrants:
   *  0 = [0,Pi/2]
   *  1 = [Pi/2,Pi]
   *  2 = [Pi,3Pi/2]
   *  3 = [3Pi/2,2Pi]
   */
  q_inf = Quadrant (x_inf);
  q_sup = Quadrant (x_sup);

  if ((q_inf == q_sup) && (BiasKSup (pX) > BiasKInf (pX) + BiasKPi)) {
    BiasKJoinRR (pR, & MinusOne, & PlusOne);
    return;
  }
  switch ((q_sup << 2) + q_inf)
    {
    case 0:
    case 3:
    case 15:
      x_inf = sin (x_inf);
      x_sup = sin (x_sup);
      y_inf = (x_inf);
      y_sup = (x_sup);
      break;
    case 1:
    case 14:
      y_inf = -1.0;
      x_inf = sin (x_inf);
      x_sup = sin (x_sup);
      x_inf = (x_inf);
      x_sup = (x_sup);
      y_sup = Maximum (x_inf, x_sup);
      break;
    case 2:
      y_inf = -1.0;
      x_sup = sin (x_sup);
      y_sup = (x_sup);
      break;
    case 4:
    case 11:
      y_sup = 1.0;
      x_inf = sin (x_inf);
      x_sup = sin (x_sup);
      x_inf = (x_inf);
      x_sup = (x_sup);
      y_inf = Minimum(x_inf, x_sup);
      break;
    case 5:
    case 9:
    case 10:
      x_inf = sin (x_inf);
      x_sup = sin (x_sup);
      y_inf = (x_sup);
      y_sup = (x_inf);
      break;
    case 6:
    case 12:
      y_inf = -1.0; y_sup = 1.0; break;
    case 7:
      y_sup = 1.0;
      x_inf = sin (x_inf);
      y_inf = (x_inf);
      break;
    case 8:
      y_sup = 1.0;
      x_sup = sin (x_sup);
      y_inf = (x_sup);
      break;
    case 13:
      y_inf = -1.0;
      x_inf = sin (x_inf);
      y_sup = (x_inf);
      break;
    }
  BiasKJoinRR (pR, & y_inf, & y_sup);
}
VOID BiasKSin(BIASKINTERVAL * const pR,
              const BIASKINTERVAL * const pX)
{
  BIASKINTERVAL proX,proR;

  BiasKPro(&proX,pX);
  BiasKSinProper(&proR,&proX);
  if(BiasKIsImproper(pX)) BiasKImp(pR,&proR);
  else (*pR)=proR;
  pR->inf = RoundDown(pR->inf);
  pR->sup = RoundUp(pR->sup);
  BiasKMeetII(pR,pR,&PlusMinusOne);
}

VOID BiasKCos (BIASKINTERVAL * const pR,
              const BIASKINTERVAL * const pX)
  /**********************************************************************
   *  R = Cos (X)
   */
{
  BIASKINTERVAL tmp;

  BiasKAddII (& tmp, pX, & BiasKPiHalfIncl);
  BiasKSin (pR, & tmp);
}

VOID BiasKTanProper (BIASKINTERVAL * const pR,
              const BIASKINTERVAL * const pX)
  /**********************************************************************
   *  R = Tan (X)
   */
{
  REAL x_inf, x_sup;
  REAL y_inf, y_sup;
  INT q_inf, q_sup;

  if (BiasKDiam (pX) >= BiasKTwoPi) {
    BiasKJoinRR (pR, & BiasKNaN, & BiasKNaN);
    return;
  }
  /* Reduction of x to [0, 2Pi] */
  ScaleTo2Pi(& x_inf, & x_sup, pX);
  if (x_sup >= x_inf + BiasKTwoPi) { /* security */
    BiasKJoinRR (pR, & BiasKNaN, & BiasKNaN);
    return;
  }
  /* x_inf, x_sup are now in the range [0, 2Pi]
   * Quadrants:
   *  0 = [0,Pi/2]
   *  1 = [Pi/2,Pi]
   *  2 = [Pi,3Pi/2]
   *  3 = [3Pi/2,2Pi]
   */
  q_inf = Quadrant (x_inf);
  q_sup = Quadrant (x_sup);

  if ((q_inf == q_sup) && (x_sup > x_inf + BiasKPi)) {
    BiasKJoinRR (pR, & BiasKNaN, & BiasKNaN);
    return;
  }
  switch ((q_sup << 2) + q_inf)
    {
    case 0:
    case 3:
    case 5:
    case 9:
    case 10:
    case 15:
      x_inf = tan (x_inf);
      x_sup = tan (x_sup);
      y_inf = (x_inf);
      y_sup = (x_sup);
      break;
    default:
      y_inf = BiasKNaN;
      y_sup = BiasKNaN;
      break;
    }
  BiasKJoinRR (pR, & y_inf, & y_sup);
}
VOID BiasKTan(BIASKINTERVAL * const pR,
              const BIASKINTERVAL * const pX)
{
  BIASKINTERVAL proX,proR;

  BiasKPro(&proX,pX);
  BiasKTanProper(&proR,&proX);
  if(BiasKIsImproper(pX)) BiasKImp(pR,&proR);
  else (*pR)=proR;
  pR->inf = RoundDown(pR->inf);
  pR->sup = RoundUp(pR->sup);
}

VOID BiasKArcSin (BIASKINTERVAL * const pR,
                 const BIASKINTERVAL * const pX)
// BiasKK
  /**********************************************************************
   *  R = ArcSin (X)
   */
{
  REAL x_inf, x_sup;

  x_inf = BiasKInf (pX); x_sup = BiasKSup (pX);
  if ((x_inf < -1.0) || (x_sup > 1.0) || (x_sup < -1.0) || (x_inf > 1.0))
    _BiasKError ("ArcSin argument out of range");
  x_inf = asin (x_inf);
  x_sup = asin (x_sup);
  x_inf = RoundDown (x_inf);
  x_sup = RoundUp   (x_sup);
  pR->inf=x_inf;
  pR->sup=x_sup;
}

VOID BiasKArcCos (BIASKINTERVAL * const pR,
                 const BIASKINTERVAL * const pX)
// BiasKK
  /**********************************************************************
   *  R = ArcCos (X)
   */
{
  REAL x_inf, x_sup;
  REAL y_inf, y_sup;

  x_inf = BiasKInf (pX); x_sup = BiasKSup (pX);
  if ((x_inf < -1.0) || (x_sup > 1.0) || (x_sup < -1.0) || (x_inf > 1.0))
    _BiasKError ("ArcCos argument out of range");
  y_inf = acos (x_sup);
  y_sup = acos (x_inf);
  y_inf = RoundUp (y_inf);
  y_sup = RoundDown   (y_sup);
  pR->inf=y_inf;
  pR->sup=y_sup;
}

VOID BiasKArcTan (BIASKINTERVAL * const pR,
                 const BIASKINTERVAL * const pX)
// BiasKK
  /**********************************************************************
   *  R = ArcTan (X)
   */
{
  REAL x_inf, x_sup;

  x_inf = atan (BiasKInf (pX));
  x_sup = atan (BiasKSup (pX));
  x_inf = RoundDown (x_inf);
  x_sup = RoundUp   (x_sup);
  pR->inf=x_inf;
  pR->sup=x_sup;
}

VOID BiasKExp (BIASKINTERVAL * const pR,
              const BIASKINTERVAL * const pX)
// BiasKK
  /**********************************************************************
   *  R = e^X
   */
{
  REAL y_inf, y_sup;

  y_inf = exp (BiasKInf (pX));
  y_sup = exp (BiasKSup (pX));
  y_inf = RoundDown (y_inf);
  y_sup = RoundUp   (y_sup);
  if (y_inf < 0.0) y_inf = 0.0;
  pR->inf=y_inf;
  pR->sup=y_sup;
}

VOID BiasKLog (BIASKINTERVAL * const pR,
              const BIASKINTERVAL * const pX)
// BiasKK
  /**********************************************************************
   *  R = Ln (X)
   */
{
  REAL y_inf, y_sup;

  if (BiasKInf (pX) <= 0.0)
    _BiasKError ("Log argument out of range"); /* BiasKHullR (pR, & BiasKNaN); */
  else {
    y_inf = log (BiasKInf (pX));
    y_sup = log (BiasKSup (pX));
    y_inf = RoundDown (y_inf);
    y_sup = RoundUp   (y_sup);
    pR->inf=y_inf;
    pR->sup=y_sup;
  }
}

VOID BiasKLog10 (BIASKINTERVAL * const pR,
                const BIASKINTERVAL * const pX)
// BiasKK
  /**********************************************************************
   *  R = Log10 (X)
   */
{
  BIASKINTERVAL t1;

  if (BiasKInf (pX) <= 0.0)
    _BiasKError ("Log10 argument out of range"); /* BiasKHullR (pR, & BiasKNaN); */
  else {
    BiasKLog (& t1, pX);
    BiasKDivII (pR, & t1, & BiasKLn10Incl);
  }
}

VOID BiasKIAbs (BIASKINTERVAL * const pR,
               const BIASKINTERVAL * const pX)
// BiasKK (no rouding performed)
  /**********************************************************************
   *  R = { |x| : x in X }
   */
{
  REAL t;

  if ((BiasKInf (pX) > 0.0)&&(BiasKSup (pX) > 0.0)) *pR = *pX;
  else if ((BiasKInf (pX) < 0.0)&&(BiasKSup (pX) < 0.0)) BiasKNeg (pR, pX);
  else {
    t = BiasKAbs (pX);
    if(BiasKIsProper(pX))
      BiasKJoinRR (pR, & Zero, & t);
    else
      BiasKMeetRR (pR, & Zero, & t);
  }
}

VOID BiasKSqr (BIASKINTERVAL * const pR,
              const BIASKINTERVAL * const pX)
// BiasKK
  /**********************************************************************
   *  R = X^2
   */
{
  BIASKINTERVAL t1;

  BiasKIAbs (& t1, pX);
  BiasKMulII (pR, & t1, & t1);
}

VOID BiasKSqrt (BIASKINTERVAL * const pR,
               const BIASKINTERVAL * const pX)
//BiasKK
  /**********************************************************************
   *  R = Sqrt (X)
   */
{
  REAL y_inf, y_sup;

  if ((BiasKInf (pX) < 0.0)||(BiasKSup (pX) < 0.0))
    _BiasKError ("Sqrt argument out of range"); /* BiasKHullR (pR, & BiasKNaN); */
  else {
    y_inf = sqrt (BiasKInf (pX));
    y_sup = sqrt (BiasKSup (pX));
    y_inf = RoundDown (y_inf);
    y_sup = RoundUp   (y_sup);
    if (y_inf < 0.0) y_inf = 0.0;
    if (y_sup < 0.0) y_sup = 0.0;
    pR->inf=y_inf;
    pR->sup=y_sup;
  }
}

VOID BiasKPowerN (BIASKINTERVAL * const pR,
                 const BIASKINTERVAL * const pX,
		 const INT n)
// BiasKK
  /**********************************************************************
   *  R = X^n
   */
{
  INT i, absn;
  BIASKINTERVAL y, z, xsqr;

  absn = (n < 0) ? (-n) : n;
  if (absn >= 2) {
    BiasKSqr (& xsqr, pX);
    if (absn & 1) BiasKMulII (& y, & xsqr, pX);
    else y = xsqr;
    for (i = 3; i < absn; i += 2) {
      z = y;
      BiasKMulII (& y, & z, & xsqr);
      }
  }
  else if (absn & 1) y = *pX;
  else BiasKSuccR2I (& y, & PlusOne);
  if (n < 0) BiasKDivRI (pR, & PlusOne, & y);
  else *pR = y;
}
